// Template for an interpose library that can be used to hook C/C++ methods
// and/or swizzle Objective-C methods for debugging/reverse-engineering.
//
// A number of methods are provided to be called from your hooks that make use
// of Apple's Symbolication framework (which though undocumented is heavily
// used by Apple utilities such as crashreporterd and dtrace).  Particularly
// useful is PrintStackTrace().
//
// Once the interpose library is built, use it as follows:
//
// A) From a Terminal prompt:
//    1) export DYLD_INSERT_LIBRARIES=/full/path/to/interpose.dylib
//    2) /path/to/application
//
// B) From gdb:
//    1) set DYLD_INSERT_LIBRARIES /full/path/to/interpose.dylib
//    2) run

#include <dlfcn.h>
#include <pthread.h>
#include <libproc.h>
#include <stdarg.h>
#include <time.h>
#include <execinfo.h>
#include <libgen.h>
#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>
#import <objc/Object.h>

class nsAutoreleasePool {
public:
    nsAutoreleasePool()
    {
        mLocalPool = [[NSAutoreleasePool alloc] init];
    }
    ~nsAutoreleasePool()
    {
        [mLocalPool release];
    }
private:
    NSAutoreleasePool *mLocalPool;
};

char *GetOwnerName(void *address);
char *GetAddressName(void *address);
void PrintAddress(void *address);
void PrintStackTrace(NSArray *addresses = nil);
BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods);

char gProcPath[PROC_PIDPATHINFO_MAXSIZE] = {0};

static void MaybeGetProcPath()
{
  if (gProcPath[0]) {
    return;
  }
  proc_pidpath(getpid(), gProcPath, sizeof(gProcPath) - 1);
}

static void LogWithFormatV(bool decorate, CFStringRef format, va_list args)
{
  MaybeGetProcPath();

  CFStringRef message = CFStringCreateWithFormatAndArguments(kCFAllocatorDefault, NULL,
                                                             format, args);

  int msgLength = CFStringGetMaximumSizeForEncoding(CFStringGetLength(message),
                                                    kCFStringEncodingUTF8);
  char *msgUTF8 = (char *) calloc(msgLength + 1, 1);
  CFStringGetCString(message, msgUTF8, msgLength, kCFStringEncodingUTF8);
  CFRelease(message);

  char *finished = (char *) calloc(msgLength + 1024, 1);
  const time_t currentTime = time(NULL);
  char timestamp[30] = {0};
  ctime_r(&currentTime, timestamp);
  timestamp[strlen(timestamp) - 1] = 0;
  if (decorate) {
    sprintf(finished, "(%s) %s[%u] %s\n",
            timestamp, gProcPath, getpid(), msgUTF8);
  } else {
    sprintf(finished, "%s\n", msgUTF8);
  }
  free(msgUTF8);

  fputs(finished, stdout);

  free(finished);
}

static void LogWithFormat(bool decorate, const char *format, ...)
{
  va_list args;
  va_start(args, format);
  CFStringRef formatCFSTR = CFStringCreateWithCString(kCFAllocatorDefault, format,
                                                      kCFStringEncodingUTF8);
  LogWithFormatV(decorate, formatCFSTR, args);
  CFRelease(formatCFSTR);
  va_end(args);
}

extern "C" void interposelib_LogWithFormatV(bool decorate, const char *format, va_list args)
{
  CFStringRef formatCFSTR = CFStringCreateWithCString(kCFAllocatorDefault, format,
                                                      kCFStringEncodingUTF8);
  LogWithFormatV(decorate, formatCFSTR, args);
  CFRelease(formatCFSTR);
}

extern "C" void interposelib_PrintStackTrace()
{
  PrintStackTrace();
}

Class gSymbolicatorClass = nil;
static BOOL gSymbolicationLoaded = NO;
static BOOL loadSymbolication()
{
  if (gSymbolicationLoaded) {
    return YES;
  }
  // Needed for printing stack traces.  Available on OS X 10.5 and up.
  if (dlopen("/System/Library/PrivateFrameworks/Symbolication.framework/Symbolication", RTLD_LAZY)) {
    gSymbolicationLoaded = YES;
    Class VMUDebugMapExtractorClass = ::NSClassFromString(@"VMUDebugMapExtractor");
    SwizzleMethods(VMUDebugMapExtractorClass, @selector(debugMapExtractorWithMachOHeader:),
                   @selector(VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:), YES);
    gSymbolicatorClass = ::NSClassFromString(@"VMUSymbolicator");

    // Swizzle other methods here
    Class NSCarbonMenuImplClass = NSClassFromString(@"NSCarbonMenuImpl");
    SwizzleMethods(NSCarbonMenuImplClass,
                   @selector(performActionWithHighlightingForItemAtIndex:),
                   @selector(NSCarbonMenuImpl_performActionWithHighlightingForItemAtIndex:),
                   NO);
  }

  return gSymbolicationLoaded;
}

pthread_t gMainThreadID = 0;

bool IsMainThread()
{
  return (gMainThreadID == pthread_self());
}

static int Hooked_pthread_once(pthread_once_t *once_control, void (*init_routine)(void))
{
  int retval = pthread_once(once_control, init_routine);
  if (!gMainThreadID) {
    gMainThreadID = pthread_self();
  }
  return retval;
}

extern "C" void *NSPushAutoreleasePool();

static void *Hooked_NSPushAutoreleasePool()
{
  void *retval = NSPushAutoreleasePool();
  if (IsMainThread()) {
    loadSymbolication();
  }
  return retval;
}

// Put other hooked methods and swizzled classes here

@interface NSObject (NSCarbonMenuImplMethodSwizzling)
- (void)NSCarbonMenuImpl_performActionWithHighlightingForItemAtIndex:(NSInteger)index;
@end

@implementation NSObject (NSCarbonMenuImplMethodSwizzling)

- (void)NSCarbonMenuImpl_performActionWithHighlightingForItemAtIndex:(NSInteger)index
{
  NSMenu *menu = (NSMenu *) [self performSelector:@selector(menu)];
  NSMenuItem *item = [menu itemAtIndex:index];
  LogWithFormat(true, "-[NSCarbonMenuImpl performActionWithHighlightingForItemAtIndex:](1), self %@, index %lli, item %@, mainMenu %@",
                self, index, item, [NSApp mainMenu]);
  [self NSCarbonMenuImpl_performActionWithHighlightingForItemAtIndex:index];
  LogWithFormat(true, "-[NSCarbonMenuImpl performActionWithHighlightingForItemAtIndex:](2), self %@, index %lli, item %@, mainMenu %@",
                self, index, item, [NSApp mainMenu]);
}

@end

typedef void (^TimerHandler)(CFRunLoopTimerRef);

class TimerInfo {
public:
  TimerInfo(CFRunLoopTimerCallBack orgCallback, void *orgUserInfo,
            CFAllocatorRetainCallBack orgUserInfoRetain,
            CFAllocatorReleaseCallBack orgUserInfoRelease)
    : mTimer(NULL), mOrgCallback(orgCallback), mOrgHandler(NULL),
      mOrgUserInfo(orgUserInfo), mOrgUserInfoRetain(orgUserInfoRetain),
      mOrgUserInfoRelease(orgUserInfoRelease), mRefCount(0)
  {
    if (mOrgUserInfo && mOrgUserInfoRetain) {
      mOrgUserInfoRetain(mOrgUserInfo);
    }
  }
  TimerInfo(TimerHandler orgHandler)
    : mTimer(NULL), mOrgCallback(NULL), mOrgHandler(NULL),
      mOrgUserInfo(NULL), mOrgUserInfoRetain(NULL),
      mOrgUserInfoRelease(NULL), mRefCount(0)
  {
    mOrgHandler = Block_copy(orgHandler);
  }
  TimerInfo(CFRunLoopTimerRef aTimer)
    : mTimer(aTimer), mOrgCallback(NULL), mOrgHandler(NULL),
      mOrgUserInfo(NULL), mOrgUserInfoRetain(NULL),
      mOrgUserInfoRelease(NULL), mRefCount(0)
  {
    CFRetain(mTimer);
  }
  ~TimerInfo()
  {
    if (mTimer) {
      CFRelease(mTimer);
    }
    if (mOrgUserInfo && mOrgUserInfoRelease) {
      mOrgUserInfoRelease(mOrgUserInfo);
    }
    if (mOrgHandler) {
      Block_release(mOrgHandler);
    }
  }

  void SetTimer(CFRunLoopTimerRef aTimer)
  {
    if (mTimer) CFRelease(mTimer);
    mTimer = aTimer;
    if (mTimer) CFRetain(mTimer);
  }

  CFRunLoopTimerRef GetTimer() { return mTimer; }
  CFRunLoopTimerCallBack GetOrgCallback() { return mOrgCallback; }
  TimerHandler GetOrgHandler() { return mOrgHandler; }
  void *GetOrgUserInfo() { return mOrgUserInfo; }

  void CountAdded() { ++mRefCount; }
  void CountRemoved() { if (--mRefCount <= 0) delete this; }
  bool NextRemoveIsLast() { return mRefCount <= 1; }

private:
  CFRunLoopTimerRef mTimer;
  CFRunLoopTimerCallBack mOrgCallback;
  TimerHandler mOrgHandler;
  void *mOrgUserInfo;
  CFAllocatorRetainCallBack mOrgUserInfoRetain;
  CFAllocatorReleaseCallBack mOrgUserInfoRelease;
  NSInteger mRefCount;
};

Boolean TimerInfoEqualCallback(const void *value1, const void *value2)
{
  TimerInfo *info1 = (TimerInfo *) value1;
  TimerInfo *info2 = (TimerInfo *) value2;
  return (info1->GetTimer() == info2->GetTimer());
}

CFHashCode TimerInfoHashCallback(const void *value)
{
  TimerInfo *info = (TimerInfo *) value;
  return CFHash(info->GetTimer());
}

CFMutableSetRef gTimerInfo = NULL;

static bool CheckTimerInfo()
{
  if (gTimerInfo) {
    return true;
  }
  CFSetCallBacks callbacks = {0};
  callbacks.equal = TimerInfoEqualCallback;
  callbacks.hash = TimerInfoHashCallback;
  gTimerInfo = CFSetCreateMutable(kCFAllocatorDefault, 0, &callbacks);
  return (gTimerInfo != NULL);
}

static void TimerCallback(CFRunLoopTimerRef timer, void *userInfo)
{
  CFStringRef mode = CFRunLoopCopyCurrentMode(CFRunLoopGetCurrent());
  TimerInfo *info = (TimerInfo *) userInfo;
  CFRunLoopTimerCallBack orgCallback = info->GetOrgCallback();
  TimerHandler orgHandler = info->GetOrgHandler();
  void *orgUserInfo = info->GetOrgUserInfo();
  if (orgCallback) {
    LogWithFormat(true, "TimerCallback(1): timer %@, userInfo %p, mode %@, time %f",
                  timer, orgUserInfo, mode, CFAbsoluteTimeGetCurrent());
    orgCallback(timer, orgUserInfo);
    LogWithFormat(true, "TimerCallback(2): timer %@, userInfo %p, mode %@, time %f",
                  timer, orgUserInfo, mode, CFAbsoluteTimeGetCurrent());
  } else {
    LogWithFormat(true, "TimerCallback(1): timer %@, mode %@, time %f, mainMenu %@",
                  timer, mode, CFAbsoluteTimeGetCurrent(), [NSApp mainMenu]);
    orgHandler(timer);
    LogWithFormat(true, "TimerCallback(2): timer %@, mode %@, time %f, mainMenu %@",
                  timer, mode, CFAbsoluteTimeGetCurrent(), [NSApp mainMenu]);
  }
  if (mode) {
    CFRelease(mode);
  }
}

static CFRunLoopTimerRef Hooked_CFRunLoopTimerCreate(CFAllocatorRef allocator,
                                                     CFAbsoluteTime fireDate,
                                                     CFTimeInterval interval,
                                                     CFOptionFlags flags,
                                                     CFIndex order,
                                                     CFRunLoopTimerCallBack callout,
                                                     CFRunLoopTimerContext *context)
{
  bool grab = false;
  const char *ownerName = "unknown";
  Dl_info addressInfo = {0};
  if (IsMainThread()) {
    void **addresses = (void **) calloc(2, sizeof(void *));
    if (addresses) {
      int count = backtrace(addresses, 2);
      if (count == 2) {
        dladdr(addresses[1], &addressInfo);
        ownerName = basename((char *)addressInfo.dli_fname);
        if (ownerName && !strcmp(ownerName, "AppKit")) {
          grab = true;
        }
      }
      free(addresses);
    }
    if (!CheckTimerInfo()) {
      grab = false;
    }
  }
  CFRunLoopTimerRef retval = NULL;
  if (grab) {
    TimerInfo *info = new TimerInfo(callout, context->info,
                                    context->retain, context->release);
    CFRunLoopTimerContext newContext = {0};
    newContext.info = info;
    retval = CFRunLoopTimerCreate(allocator, fireDate,
                                  interval, flags, order,
                                  TimerCallback, &newContext);
    if (retval) {
      info->SetTimer(retval);
      CFSetAddValue(gTimerInfo, info);
    } else {
      delete info;
    }
  } else {
    retval = CFRunLoopTimerCreate(allocator, fireDate,
                                  interval, flags, order,
                                  callout, context);
  }
  if (grab) {
    LogWithFormat(true, "CFRunLoopTimerCreate(): owner %s, returning %@",
                  ownerName, retval);
  }
  return retval;
}

static CFRunLoopTimerRef Hooked_CFRunLoopTimerCreateWithHandler(CFAllocatorRef allocator,
                                                                CFAbsoluteTime fireDate,
                                                                CFTimeInterval interval,
                                                                CFOptionFlags flags,
                                                                CFIndex order,
                                                                void (^block)(CFRunLoopTimerRef timer))
{
  bool grab = false;
  const char *ownerName = "unknown";
  Dl_info addressInfo = {0};
  if (IsMainThread()) {
    void **addresses = (void **) calloc(2, sizeof(void *));
    if (addresses) {
      int count = backtrace(addresses, 2);
      if (count == 2) {
        dladdr(addresses[1], &addressInfo);
        ownerName = basename((char *)addressInfo.dli_fname);
        if (ownerName && !strcmp(ownerName, "AppKit")) {
          grab = true;
        }
      }
      free(addresses);
    }
    if (!CheckTimerInfo()) {
      grab = false;
    }
  }
  CFRunLoopTimerRef retval = NULL;
  if (grab) {
    TimerInfo *info = new TimerInfo(block);
    CFRunLoopTimerContext newContext = {0};
    newContext.info = info;
    retval = CFRunLoopTimerCreate(allocator, fireDate,
                                  interval, flags, order,
                                  TimerCallback, &newContext);
    if (retval) {
      info->SetTimer(retval);
      CFSetAddValue(gTimerInfo, info);
    } else {
      delete info;
    }
  } else {
    retval = CFRunLoopTimerCreateWithHandler(allocator, fireDate,
                                             interval, flags, order,
                                             block);
  }
  if (grab) {
    LogWithFormat(true, "CFRunLoopTimerCreateWithHandler(): owner %s, returning %@",
                  ownerName, retval);
  }
  return retval;
}

static void Hooked_CFRunLoopAddTimer(CFRunLoopRef rl, CFRunLoopTimerRef timer,
                                     CFStringRef mode)
{
  bool grab = false;
  const char *ownerName = "unknown";
  Dl_info addressInfo = {0};
  if (IsMainThread()) {
    void **addresses = (void **) calloc(2, sizeof(void *));
    if (addresses) {
      int count = backtrace(addresses, 2);
      if (count == 2) {
        dladdr(addresses[1], &addressInfo);
        ownerName = basename((char *)addressInfo.dli_fname);
        if (ownerName && !strcmp(ownerName, "AppKit")) {
          grab = true;
        }
      }
      free(addresses);
    }
    if (!CheckTimerInfo()) {
      grab = false;
    }
  }
  TimerInfo info(timer);
  if (grab) {
    grab = CFSetContainsValue(gTimerInfo, &info);
  }
  if (grab) {
    TimerInfo *orgInfo = (TimerInfo *) CFSetGetValue(gTimerInfo, &info);
    orgInfo->CountAdded();
  }
  CFRunLoopAddTimer(rl, timer, mode);
  if (grab) {
    LogWithFormat(true, "CFRunLoopAddTimer(): timer %@, mode %@, owner %s",
                  timer, mode, ownerName);
  }
}

static void Hooked_CFRunLoopRemoveTimer(CFRunLoopRef rl, CFRunLoopTimerRef timer,
                                        CFStringRef mode)
{
  bool grab = false;
  const char *ownerName = "unknown";
  Dl_info addressInfo = {0};
  if (IsMainThread()) {
    void **addresses = (void **) calloc(2, sizeof(void *));
    if (addresses) {
      int count = backtrace(addresses, 2);
      if (count == 2) {
        dladdr(addresses[1], &addressInfo);
        ownerName = basename((char *)addressInfo.dli_fname);
        if (ownerName && !strcmp(ownerName, "AppKit")) {
          grab = true;
        }
      }
      free(addresses);
    }
    if (!CheckTimerInfo()) {
      grab = false;
    }
  }
  TimerInfo info(timer);
  if (grab) {
    grab = CFSetContainsValue(gTimerInfo, &info);
  }
  if (grab) {
    TimerInfo *orgInfo = (TimerInfo *) CFSetGetValue(gTimerInfo, &info);
    if (orgInfo->NextRemoveIsLast()) {
      CFSetRemoveValue(gTimerInfo, orgInfo);
    }
    orgInfo->CountRemoved();
  }
  if (grab) {
    LogWithFormat(true, "CFRunLoopRemoveTimer(): timer %@, mode %@, owner %s",
                  timer, mode, ownerName);
  }
  CFRunLoopRemoveTimer(rl, timer, mode);
}

static void Hooked_CFRunLoopTimerInvalidate(CFRunLoopTimerRef timer)
{
  bool grab = false;
  const char *ownerName = "unknown";
  Dl_info addressInfo = {0};
  if (IsMainThread()) {
    void **addresses = (void **) calloc(2, sizeof(void *));
    if (addresses) {
      int count = backtrace(addresses, 2);
      if (count == 2) {
        dladdr(addresses[1], &addressInfo);
        ownerName = basename((char *)addressInfo.dli_fname);
        if (ownerName && !strcmp(ownerName, "AppKit")) {
          grab = true;
        }
      }
      free(addresses);
    }
    if (!CheckTimerInfo()) {
      grab = false;
    }
  }
  TimerInfo info(timer);
  if (grab) {
    grab = CFSetContainsValue(gTimerInfo, &info);
  }
  if (grab) {
    TimerInfo *orgInfo = (TimerInfo *) CFSetGetValue(gTimerInfo, &info);
    CFSetRemoveValue(gTimerInfo, orgInfo);
    delete orgInfo;
  }
  if (grab) {
    LogWithFormat(true, "CFRunLoopTimerInvalidate(): timer %@, owner %s",
                  timer, ownerName);
  }
  CFRunLoopTimerInvalidate(timer);
}

extern "C" void HiliteMenu(MenuID menuID);
extern "C" MenuRef GetMenuHandle(MenuID menuID);

static void Hooked_HiliteMenu(MenuID menuID)
{
  HiliteMenu(menuID);
  LogWithFormat(true, "HiliteMenu(): menuID %p", menuID);
}

#pragma mark -

struct interpose_substitution {
  const void* replacement;
  const void* original;
};

#define INTERPOSE_FUNCTION(function) \
    { reinterpret_cast<const void*>(Hooked_##function), \
      reinterpret_cast<const void*>(function) }

__attribute__((used)) static const interpose_substitution substitutions[]
    __attribute__((section("__DATA, __interpose"))) = {
  INTERPOSE_FUNCTION(pthread_once),
  INTERPOSE_FUNCTION(NSPushAutoreleasePool),
  //INTERPOSE_FUNCTION(CFRunLoopTimerCreate),
  INTERPOSE_FUNCTION(CFRunLoopTimerCreateWithHandler),
  INTERPOSE_FUNCTION(CFRunLoopAddTimer),
  INTERPOSE_FUNCTION(CFRunLoopRemoveTimer),
  INTERPOSE_FUNCTION(CFRunLoopTimerInvalidate),
  INTERPOSE_FUNCTION(HiliteMenu),
};

// What follows are definitions of the Symbolication interfaces we use to get
// stack traces.  The Symbolication framework is an undocumented, private
// framework available on OS X 10.5 and up.  It's used by Apple utilities
// such as dtrace.  http://www.cocoadev.com/index.pl?StackTraces has basic
// information on how to use the Symbolication framework.

@interface VMUAddressRange : NSObject <NSCoding>
{
}

@end

@interface VMUSymbol : VMUAddressRange <NSCopying>
{
}

- (NSString *)name;
- (NSString *)description;
- (id)text;

@end

@interface VMUSymbolOwner : NSObject <NSCopying>
{
}

- (NSString *)name;

@end

// We must override the debugMapExtractorWithMachOHeader method:  Otherwise
// the Symbolication framework can take a minute or more to initialize!!
// We don't need the information in the "debug maps" (Dwarf or dsym debugging
// information) to get simple stack traces.
@interface NSObject (VMUDebugMapExtractorMethodSwizzling)
+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader;
@end

@implementation NSObject (VMUDebugMapExtractorMethodSwizzling)

+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader
{
  return nil;
}

@end

@interface VMUSymbolicator : NSObject
{
}

+ (VMUSymbolicator *)symbolicatorForPid:(int)pid;
- (VMUSymbol *)symbolForAddress:(unsigned long long)address;
- (VMUSymbolOwner *)symbolOwnerForAddress:(unsigned long long)address;

@end

@interface VMUSymbolicator (PerformanceOptimization)
- (id)symbolOwnerNameForAddress:(unsigned long long)fp8;
@end

char *GetOwnerName(void *address)
{
  static char holder[1024] = {0};
  if (!loadSymbolication()) {
    holder[0] = 0;
    return holder;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    holder[0] = 0;
    return holder;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  NSString *ownerName = nil;
  if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
    ownerName =
      [symbolicator symbolOwnerNameForAddress:(unsigned long long)address];
  } else {
    VMUSymbolOwner *owner =
      [symbolicator symbolOwnerForAddress:(unsigned long long)address];
    ownerName = [owner name];
  }
  if (ownerName) {
    strcpy(holder, [ownerName UTF8String]);
    return holder;
  }
  holder[0] = 0;
  return holder;
}

char *GetAddressName(void *address)
{
  static char holder[1024] = {0};
  if (!loadSymbolication()) {
    holder[0] = 0;
    return holder;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    holder[0] = 0;
    return holder;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  if (symbol && [symbol name]) {
    strcpy(holder, [[symbol name] UTF8String]);
    return holder;
  }
  holder[0] = 0;
  return holder;
}

void PrintAddress(void *address)
{
  if (!loadSymbolication()) {
    return;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    return;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  NSString *ownerName = nil;
  if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
    ownerName =
      [symbolicator symbolOwnerNameForAddress:(unsigned long long)address];
  } else {
    VMUSymbolOwner *owner =
      [symbolicator symbolOwnerForAddress:(unsigned long long)address];
    ownerName = [owner name];
  }
  LogWithFormat(false, "    (%@) %@, %p", ownerName, [symbol name], address);
}

void PrintStackTrace(NSArray *addresses)
{
  if (!loadSymbolication()) {
    return;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    return;
  }

  if (!addresses) {
    addresses = [NSThread callStackReturnAddresses];
  }
  NSUInteger count = [addresses count];
  for (NSUInteger i = 0; i < count; ++i) {
    NSNumber *item = (NSNumber *) [addresses objectAtIndex:i];
    VMUSymbol *symbol =
      [symbolicator symbolForAddress:[item unsignedLongLongValue]];
    NSString *ownerName = nil;
    if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
      ownerName =
        [symbolicator symbolOwnerNameForAddress:[item unsignedLongLongValue]];
    } else {
      VMUSymbolOwner *owner =
        [symbolicator symbolOwnerForAddress:[item unsignedLongLongValue]];
      ownerName = [owner name];
    }
    LogWithFormat(false, "    (%@) %@, %p",
                  ownerName, [symbol name], (void *)[item unsignedLongLongValue]);
  }
}

BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods)
{
  Method original = nil;
  Method posed = nil;

  if (classMethods) {
    original = class_getClassMethod(aClass, orgMethod);
    posed = class_getClassMethod(aClass, posedMethod);
  } else {
    original = class_getInstanceMethod(aClass, orgMethod);
    posed = class_getInstanceMethod(aClass, posedMethod);
  }

  if (!original || !posed)
    return NO;

  method_exchangeImplementations(original, posed);

  return YES;
}
